' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_LmmVm_002.spin                                                                           *
' *                                                                                                     *
' *******************************************************************************************************

PRI Version
  return String("AiChip_LmmVm_002")
  
CON
{

Changes to AiChip_LmmVm_001

  Reduced code size of LMM_Djnz, LMM_Tjz and LMM_Tjnz at the cost of slightly slower execution.

  Added LMM_Gti, LMM_Sti and LMM_Execute to allow register indirect loads and stores and opcodes to be
  built and executed.

  Added LMM_Case to allowed indexed jumps.
  
  Added LMM_Skip to complement LMM_Conditional.

  Various cosmetic and documentation changes.
  
}
CON
{

This is a Virtual Machine (VM) which allows execution of Large Memory Model (LMM) programs. LMM programs
are those which can be larger than the 496 longs allowed for by a Cog's memory limitation. This VM is
designed to be easy to use with the Propeller Tool although it does necessitate some manual changes to
use a Propeller Assembler program as an LMM program.

This VM only supports user variables which are held within the executing Cog memory and the number of
variables is an absolute maximum of 496 and will be lower than that because of the space taken up by
the VM code itself. If the internal stack is used, the number of user variables is reduced further.

The current LMM VM footprint is approximately 72 longs ( excluding debugging code ) and can be reduced
to around 60 longs if conditional LMM instructions are not required ( LMM_Conditional removed ). There
are therefore around 420 Cog memory locations available for user registers and internal stack.

The following changes must be made to make a Propeller Assembly program LMM compatible -

                mov     <reg>,<longConstant>            jmp     #LMM_Load
                                                        long    <reg>
                                                        long    <longConstant>

                jmp     #<Label>                        jmp     #LMM_Jmp
                                                        long    @<Label>

                jmp     <reg>                           jmp     #LMM_Jmp_Reg
                                                        long    <reg>

                call    #<Label>                        jmp     #LMM_Call
                                                        long    @<Label>

                jmpret  <Label>_Ret,#<Label>            jmp     #LMM_Call
                                                        long    @<Label>

                ret                                     jmp     #LMM_Ret

                djnz    <reg>,#<Label>                  jmp     #LMM_Djnz
                                                        long    <reg>
                                                        long    @<Label>

                tjz     <reg>,#<Label>                  jmp     #LMM_Tjz
                                                        long    <reg>
                                                        long    @<Label>

                tjnz    <reg>,#<Label>                  jmp     #LMM_Tjnz
                                                        long    <reg>
                                                        long    @<Label>
         
If any of the above need to be conditionally executed they should be preceded by 'jmp #LMM_Conditional'
as in the following example ...

        IF_Z    call    #<Label>                        jmp     #LMM_Conditional
                                                IF_Z    jmp     #LMM_Call
                                                        long    @<Label>

Additionally, a 'jmp #LMM' can be skipped by using 'jmp #LMM_Skip' ...

        IF_Z    call    #<Label>                IF_NZ   jmp     #LMM_Skip
                                                        jmp     #LMM_Call
                                                        long    @<Label>

Do not precede a 'jmp #LMM_Skip' with a 'jmp #LMM_Conditional' or vice-versa as the VM is not coded
to handle that. If it is required, LMM_Skip will need to be modified. Only put a 'jmp #LMM_Conditional'
and 'jmp #LMM_Skip' before another 'jmp #LMM' instruction, not a native Propeller instrcution. Note
that 'jmp #LMM_COnditional' and 'jmp #LMM_Skip' cannot be used before 'jmp #LMM_Case'.

To reduce the size of the VM, LMM_Conditional should be removed and only LMM_Skip used.

The following Propeller Assembly instructions are not currently supported for LMM programs and must
not be used -

                jmpret  reg,reg
                djnz    reg,reg
                tjz     reg,reg
                tjnz    reg,reg

This VM uses a software stack rather than modifying 'ret' instructions as a genuine Propeller Cog
program would. An internal stack is used by default which grows down in Cog memory from $1EF. When
a large number of user Registers are defined care should be taken to avoid stack overflow corrupting
registers, or an external stack may be used. If an external stack is required it should be enabled
by using -

                jmp     #LMM_Load
                long    sp
                long    <stackAddress>

When an external stack is used, the stack will grow upwards in memory from the start of the array. To
switch back from the external stack to the internal stack, use -

                jmp     #LMM_Load
                long    sp
                long    0

Stack swithing between internal and external can be done dynamically as required. Take care to return
from the same stack as any call was issued upon. There is little, if any, discernable deterioration
in performnce when using an external ( hub memory ) stack as opposed to the internal stack.

Calls into routines added within the LMM or placed in the LMM Program's Register Area can be called
with the traditional 'jmp #Label' and will execute full speed then continue LMM execution at the LMM
instruction immediately after the call. Such calls use the traditional 'update ret' mechanism and do
not use any stack, internal or in hub memory.

Note that when converting 'jmp reg' the 'reg' should have been loaded with an '@Label' of where the
destination jump will be to. For example, the following are equivalent -

                jmp     #LMM_Jmp                        jmp     #LMM_Load
                long    @<Label>                        long    <reg>
                                                        long    @<Label>
                                                        :
                                                        jmp     #LMM_Jmp_Reg
                                                        long    <reg>

To enable register indirect operations and dynamically building opcodes to execute, three additional
LMM functions have been added -

Get Register Indirect - mov dreg,[sreg]

                movs    :Src,<sreg>                     jmp     #LMM_Gti
                :                                       long    <dreg>
        :Src    mov     <dreg>,0-0                      long    <sreg>

Store Register Indirect - mov [dreg],sreg

                movd    :Dst,<dreg>                     jmp     #LMM_Sti
                :                                       long    <dreg>
        :Dst    mov     0-0,<sreg>                      long    <sreg>

Execute Opcode                                          

                movi    :Opc,????                       movi    <opc>,????
                movd    :Opc,????                       movd    <opc>,????
                movs    :Opc,????                       movs    <opc>,????
                :                                       jmp     #LMM_Execute
        :Opc    long    0-0                             long    <opc>

To allow for easy indexed jumping and to replace indexed lookup and jmp operations a 'jmp LMM_Case'
instruction has been added. A register contains an index 0-N and the 'jmp #LMM_Case' contains a list
of addresses to jump to. Note the addresses are specified using 'word' not 'long' ...

                jmp     #LMM_Case
                long    reg
                word    @Label0
                word    @Label1
                :
                word    @LabelN

Note that 'jmp #LMM_Case' cannot be preceded by 'jmp #LMM_Conditional' nor 'jmp #LMM_Conditional'.

The current LMM substitutions for Propeller instructions have been chosen to be easy to enter by hand
and use within the Propeller Tool and these could be optimised. The 'jmp #LMM' commands can have an
associated register embedded within the command ( as the unused destination register ) and two word
arguments can be compacted as a single long. The consequence though is that any increased code density
will be offset by decreased speed of execution.

There is no overlay or 'block load and execute' capability in this VM. The VM is however very easily
extensible to add overlays or other additional VM instructions such as Push and Pop register.

LMM Instructions tested ...

  LMM_Load
  LMM_Jmp
  LMM_Jmp_Reg
  LMM_Call
  LMM_Ret

LMM Instructions needing re-testing ...

  LMM_Djnz
  LMM_Tjz
  LMM_Tjnz

LMM Instructions not tested ...

  LMM_Gti
  LMM_Sti
  LMM_Execute
  LMM_Case
  LMM_Conditional
  LMM_Skip

}
CON
{
  
PLEASE NOTE : If modifying and re-releasing this code, please change the prefix of the source file from
              AiChip so as to avoid any confusion as to the origin of the source and update all internal
              references as appropriate. If you wish to credit the orginal source, anything  along the
              lines of, "Based on original AiChip_LmmVm_XXX from AiChip Industries", is acceptable.

              AiChip is a trademark of AiChip Industries.

}
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE      = XTAL1 + PLL16x                        ' 80MHz System Clock Speed
  _XINFREQ      = 5_000_000

OBJ

  tv            : "TV_Text"                             ' TV Text Display used for debugging

CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  TV_PIN        = 12                                    ' Base TV Pin

VAR

  long stack[1024]                                      ' Stack area to use for Spin test programs
  long cogDump[$200]                                    ' Copy of cog memory when debug event occurs

PUB Main

  tv.Start( TV_PIN )
  tv.Str( Version )

  Case 2
  
    0 : DebugLmm                                        ' Benchmark Timing      mm:ss   sec  ratios
                                                        
    1 : CogNew( SpinTest1, @stack )                     ' 1 - Spin timing test  15:40   940   1:47
    2 : CogNew( @LmmVm, @LmmTest1 )                     ' 2 - LMM  timing test  02:00   120   1:6
    3 : CogNew( @PasmTest1, 0 )                         ' 3 - PASM timing test  00:20    20   1:1

    4 : CogNew( SpinTest2, @stack )                     ' 4 - Spin timing test  28:00  1680   1:42
    5 : CogNew( @LmmVm, @LmmTest2 )                     ' 5 - LMM  timing test  04:20   260   1:6.5
    6 : CogNew( @PasmTest2, 0 )                         ' 6 - PASM timing test  00:40    40   1:1

  repeat
    tv.Out( $01 )
    tv.Out( $0D )
    tv.Dec( long[ $7000 ] )
    if long[ $7000 ] => 100_000_000
      tv.Str( String(" FINISHED") )

PRI DebugLmm | event, cogAdr

  CogNew( @LmmVm, @LmmTest1 )

  repeat
    repeat while ( event := long[ $7FF0 ] ) == 0        ' Wait until Lmm VM talks back
    tv.Out( $0D )
    repeat cogAdr from $000 to $1FF
      long[ $7FF0 ] := 0
      repeat while long[ $7FF0] == 0
      cogDump[ cogAdr ] := long[ $7FEC ]
    case event
      1     : IsDebug                                   ' 1 Debug
      2     : IsDebug                                   ' 2 Stop
              tv.Str( String($0D,"Stop") )
              repeat
      3     : IsSingleStep                              ' 3 SingleStep
      4     : IsNotImplemented                          ' 4 NotImplemented
      other : tv.Str( String("Unknown ") )              ' ? Unknown
              tv.Hex( event, 8 )
              repeat
    long[ $7FEC ] := 0                                  ' Let the VM execute another instruction
    long[ $7FF0 ] := 0

PRI IsDebug

  tv.Out( "+")
  tv.Hex( ( Cog(@Debug_Ret) & $1FF ) - 1, 3 )
  tv.Out( " " )
  tv.Hex( long[ $7FF4 ], 8 )
  tv.Out( " " )
  tv.Hex( long[ $7FF8 ], 8 )
  tv.Out( " " )
  tv.Hex( long[ $7FFC ], 8 )
  PauseMs( 1000 )
  
PRI IsSingleStep | xPc, xOp

  xPc := Cog(@pc)
  xOp := long[ xPc ]
  tv.Hex( xPc, 4 )
  tv.str( String(" IF ") )
  case ( xOp >> 18 ) & $F
    $0 : tv.str( String("NEVER  ") )
    $1 : tv.str( String("NC & NZ") )
    $2 : tv.str( String("NC & Z ") )
    $3 : tv.str( String("NC     ") )
    $4 : tv.str( String("C & NZ ") )
    $5 : tv.str( String("NZ     ") )
    $6 : tv.str( String("C <> Z ") )
    $7 : tv.str( String("NC | NZ") )
    $8 : tv.str( String("C & Z  ") )
    $9 : tv.str( String("C == Z ") )
    $A : tv.str( String("Z      ") )
    $B : tv.str( String("NC | Z ") )
    $C : tv.str( String("C      ") )
    $D : tv.str( String("C | NZ ") )
    $E : tv.str( String("C | Z  ") )
    $F : tv.str( String("ALWAYS ") )
  tv.out( " " )
  tv.bin( xOp >> 26, 6 )
  tv.out( " " )
  tv.out( "$" )
  tv.hex( ( xOp >> 9 ) & $1FF, 3 )
  tv.out( "," )
  if ( xOp >> 22) & 1
    tv.out( "#" )
  tv.out( "$" )
  tv.hex( xOp & $1FF, 3 )
  if ( ( xOp >> 23) & 1 ) == 0
    tv.str( String(" NR") )
  if ( xOp >> 24) & 1
    tv.str( String(" WC") )
  if ( xOp >> 25) & 1
    tv.str( String(" WZ") )
  PauseMs( 1000 )

PRI IsNotImplemented

  tv.Str( String("Not Implemented before $") )
  tv.Hex( Cog(@pc), 4 )
  repeat

PRI Cog( AtCogAddress )
  result := cogDump[ ( AtCogAddress - @LmmVm ) >> 2 ]

PRI PauseMs( mS )
  waitcnt( CLKFREQ / 1000 * mS + CNT )
  
CON

' *******************************************************************************************************
' *                                                                                                     *
' *             Spin and PASM Timing Test Programs                                                      *
' *                                                                                                     *
' *******************************************************************************************************

PRI SpinTest1

  repeat
    ++long[ $7000 ]

PRI SpinTest2

  repeat
    SpinTest2Sub

PRI SpinTest2Sub

    ++long[ $7000 ]

DAT             org     $000

PasmTest1       add     pasmVar1,#1
                wrlong  pasmVar1,k1_0000_7000
                jmp     #PasmTest1

PasmVar1        long    0         

k1_0000_7000    long    $0000_7000

DAT             org     $000

PasmTest2       call   #PasmSub2
                jmp    #PasmTest2

PasmSub2        add     pasmVar2,#1
                wrlong  pasmVar2,k2_0000_7000
PasmSub2_Ret    ret

PasmVar2        long    0         

k2_0000_7000    long    $0000_7000

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *             LMM Interpreter                                                                         *
' *                                                                                                     *
' *******************************************************************************************************

LmmVm           org     $000

' .-----------------------------------------------------------------------------------------------------.
' |             Initialisation / Re-used for Internal Registers                                         |
' `-----------------------------------------------------------------------------------------------------'

pc              mov     pc,PAR
acc             jmp     #LMM_Fetch

' *******************************************************************************************************
' *                                                                                                     *
' *     Debugging Utilities                                                                             *
' *                                                                                                     *
' *******************************************************************************************************
'
' Debugging uses the top 20 bytes ( 5 longs ) of Ram ( $7FEC to $7FFF ) to communicate with the Spin
' wrapper.
' 
' These routines allow the cog to inform the Spin wrapper what's going on within itself. In all cases 
' the long[$7FF0] is set with a flag to indicate what internal 'event' is being generated. Then, once
' long[$7FF0] has been cleared, the entire $200 longs of cog are passed to the Spin program a long at 
' a time. This is done by putting the cog value in long[$7FEC] and setting long[$7FF0]. Every time the
' long[$7FF0] is cleared by the Spin wrapper the next long is passed out. After $200 have been passed
' over the code here waits for long[$7FF0] to be cleared again and execution continues. It is up to the
' Spin wrapper to not allow the cog to continue after a 'Stop' or 'NotImplemented' event.
'
' Note that during the copy we load 'dbgTmp' then wrlong that rather than altering the destination of
' the wrlong itself; this is so we can read the SPR's which cannot be used as destination registers. It
' also helps with the counting method used.
'
' In addition, writes to long[$7FF4], long[$7FF8] and long[$7FFC] can be used to pass information back
' to the debugger prior to calling any of the debugging routines. These are displayed on Debug and Stop
' calls. The Spin wrapper dictates how the events are handled and what is displayed.
'
' Example ...
'
'               wrlong  lhs,k_0000_7FF4
'               wrlong  rhs,k_0000_7FF8
'               wrlong  acc,k_0000_7FFC
'               call    #Debug
'
' Note we have to protect the Z flag so we don't screw any Cog code which relies on it being unchanged
' over calls to Debug etc. We do that by using the lsb of the destination register of the Debug_Ret
' which is otherwise unused. It saves a long. We don't need to protect C because we don't use WC in the
' debugger. 

NotImplemented  add    dbgTmp,#1                        ' 4
SingleStep      add    dbgTmp,#1                        ' 3
Stop            add    dbgTmp,#1                        ' 2
Debug           add    dbgTmp,#1                        ' 1

                wrlong  dbgTmp,k_0000_7FF0

                muxnz   Debug_Ret,k_0000_0200           ' Protect Z flag

:Dump           call    #DebugWait
:Src            mov     dbgTmp,$000                     ' No need to pre-initialise this
                wrlong  dbgTmp,k_0000_7FEC
                add     :Src,#1
                and     :Src,#$1FF WZ, NR               ' If we hit $200 we corrupted the opcode
        IF_Z    sub     :Src,k_0000_0200                ' So set it to $000 ...
                wrlong  k_0000_7FF0,k_0000_7FF0
        IF_NZ   jmp     #:Dump                          ' ... and no more to send out

                call    #DebugWait                      ' leaves dbgTmp zeroed on exit

                and     Debug_Ret,k_0000_0200 WZ, NR    ' Restore Z flag
                
NotImplemented_Ret
SingleStep_Ret
Stop_Ret
Debug_Ret       ret

DebugWait       rdlong  dbgTmp,k_0000_7FF0
                tjnz    dbgTmp,#DebugWait
DebugWait_Ret   ret

dbgTmp          long    0

k_0000_7FEC     long    $0000_7FEC
k_0000_7FF0     long    $0000_7FF0
k_0000_7FF4     long    $0000_7FF4
k_0000_7FF8     long    $0000_7FF8
k_0000_7FFC     long    $0000_7FFC

DAT

' .-----------------------------------------------------------------------------------------------------.
' |             LMM Internal registers and constants                                                    |
' `-----------------------------------------------------------------------------------------------------'

k_CogBp_Minus_4 long    $0010-4                         ' Base adjust for @Label to  hub memory address

k_0000_0200     long    $0000_0200                      ' ---- ---- ---- --dd dddd ddD- ---- ----
k_003C_0000     long    $003C_0000                      ' oooo oozc riCC CCdd dddd ddds ssss ssss

sp              long    0                               ' Stack Pointer ( 0 = Internal Stack )

' .-----------------------------------------------------------------------------------------------------.
' |             JMP     #@Label                         jmp     #LMM_Jmp                                |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Add4_Jmp    add     pc,#4

LMM_Jmp         rdlong  pc,pc

' .-----------------------------------------------------------------------------------------------------.
' |             Core Fetch and Execute Loop                                                             |
' `-----------------------------------------------------------------------------------------------------'

LMM_AddX_Fetch  add     pc,k_cogBp_Minus_4              ' Adjust @Label address to actual hub addresses

LMM_Add4_Fetch  add     pc,#4

LMM_Fetch ' === call    #SingleStep                     ' Add this in for debugging tracing

                rdlong  :Opc,pc
                add     pc,#4
:Opc            nop
                jmp     #LMM_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             LOAD    reg,#longConstant               jmp     #LMM_Load                               |
' |                                                     long    reg                                     |
' |                                                     long    longConstant                            |
' `-----------------------------------------------------------------------------------------------------'

LMM_Load        rdlong  acc,pc
                movd    :Dst,acc
                add     pc,#4
:Dst            rdlong  0-0,pc
                jmp     #LMM_Add4_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             GTI     dreg,sreg                       jmp     #LMM_Gti                                |
' |             MOV     dreg,[sreg]                     long    dreg                                    |
' |                                                     long    sreg                                    |
' `-----------------------------------------------------------------------------------------------------'

LMM_Gti         rdlong  acc,pc
                movd    LMM_Gti_DstSrc,acc
                add     pc,#4
                rdlong  acc,pc
                movs    :Src,acc
                nop           
:Src            movs    LMM_Gti_DstSrc,0-0

LMM_Gti_Xxx     add     pc,#4
LMM_Gti_DstSrc  mov     0-0,0-0
                jmp     #LMM_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             STI     dreg,sreg                       jmp     #LMM_Sti                                |
' |             MOV     [dreg],sreg                     long    dreg                                    |
' |                                                     long    sreg                                    |
' `-----------------------------------------------------------------------------------------------------'

LMM_Sti         rdlong  acc,pc
                movs    :Src,acc
                add     pc,#4
                rdlong  acc,pc
                movs    LMM_Gti_DstSrc,acc
:Src            movd    LMM_Gti_DstSrc,0-0
                
                jmp     #LMM_Gti_Xxx

' .-----------------------------------------------------------------------------------------------------.
' |             EXECUTE reg                             jmp     #LMM_Execute                            |
' |                                                     long    reg                                     |
' `-----------------------------------------------------------------------------------------------------'

LMM_Execute     rdlong  acc,pc
                mov     :Opc,acc
                add     pc,#4
:Opc            long    0-0
                jmp     #LMM_Fetch
                
' .-----------------------------------------------------------------------------------------------------.
' |             JMP     reg                             jmp     #LMM_Jmp_Reg                            |
' |                                                     long    reg                                     |
' `-----------------------------------------------------------------------------------------------------'

LMM_Jmp_Reg     rdlong  acc,pc
                movs    :Src,acc
                nop
:Src            mov     pc,0-0
                jmp     #LMM_AddX_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             CASE     reg                            jmp     #LMM_Case                               |
' |                                                     long    reg                                     |
' |                                                     word    @Label0                                 |
' |                                                     word    @Label1                                 |
' |                                                     :                                               |
' |                                                     word    @LabelN                                 |
' `-----------------------------------------------------------------------------------------------------'

LMM_Case        rdlong  acc,pc
                add     pc,#4
                shl     acc,#1
                add     pc,acc
                rdword  pc,pc
                jmp     #LMM_Fetch
'
' .-----------------------------------------------------------------------------------------------------.
' |             CALL    #@Label                         jmp     #LMM_Call                               |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Call        tjz     sp,#LMM_Call_Int                ' Use internal stack if sp == 0

                wrword  pc,sp                           ' Push to external stack
                add     sp,#2
                jmp     #LMM_Jmp
                
LMM_Call_Int    mov     $1EF,pc                         ' Push to internal stack
                sub     LMM_Call_Int,k_0000_0200        ' Adjust push address
                sub     LMM_Ret_Int,#1                  ' Adjust pop address
                jmp     #LMM_Jmp

' .-----------------------------------------------------------------------------------------------------.
' |             RET                                     jmp     #LMM_Ret                                |
' `-----------------------------------------------------------------------------------------------------'

LMM_Ret         tjz     sp,#LMM_Ret_Int                 ' Use internal stack if sp == 0

                sub     sp,#2                           ' Pop from external stack
                rdword  pc,sp
                jmp     #LMM_Add4_Fetch
                                                        
LMM_Ret_Int     mov     pc,$1F0                         ' Pop from internal stack
                add     LMM_Call_Int,k_0000_0200        ' Adjust push address
                add     LMM_Ret_Int,#1                  ' Adjust pop address
                jmp     #LMM_Add4_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             DJNZ    reg,#@Label                     jmp     #LMM_Djnz                               |
' |                                                     long    reg                                     |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Djnz        movi    LMM_Djnz_OpcDst,#%111001_00_1   ' Djnz opcode

LMM_Djnz_Xxx    rdlong  acc,pc
                movd    LMM_Djnz_OpcDst,acc
                add     pc,#4
LMM_Djnz_OpcDst djnz    0-0,#LMM_Jmp
                jmp     #LMM_Add4_Fetch

' .-----------------------------------------------------------------------------------------------------.
' |             TJZ     reg,#@Label                     jmp     #LMM_Tjz                                |
' |                                                     long    reg                                     |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Tjz         movi    LMM_Djnz_OpcDst,#%111011_00_0   ' Tjz opcode
                jmp     #LMM_Djnz_Xxx

' .-----------------------------------------------------------------------------------------------------.
' |             TJNZ    reg,#@Label                     jmp     #LMM_Tjnz                               |
' |                                                     long    reg                                     |
' |                                                     long    @Label                                  |
' `-----------------------------------------------------------------------------------------------------'

LMM_Tjnz        movi    LMM_Djnz_OpcDst,#%111010_00_0   ' Tjnz Opcode
                jmp     #LMM_Djnz_Xxx

' .-----------------------------------------------------------------------------------------------------.
' |             CONDITIONAL                             jmp     #LMM_Conditional                        |
' `-----------------------------------------------------------------------------------------------------'

LMM_Conditional rdlong  acc,pc                          ' Get the condition code of the next 'jmp #LMM'
                and     acc,k_003C_0000                 ' oooo oozc riCC CCdd dddd ddds ssss ssss
                andn    :Jmp,k_003C_0000                ' Modify the jmp at :Jmp so it is conditional
                or      :Jmp,acc
                nop
:Jmp            jmp     #LMM_Fetch                      ' If passes condition, execute what follows
                                                        ' Otherwise fall through and skip it ...
                                                        
' .-----------------------------------------------------------------------------------------------------.
' |             SKIP                                    jmp     #LMM_Skip                               |
' `-----------------------------------------------------------------------------------------------------'

' Skip the following 'jmp #LMM and all the parameters which follow it. We do it this way to avoid
' altering the Z and C bits or they need storing and restoring. It's slow, but it is hoped there will be
' few conditional 'jmp #LMM' calls needed in any LMM program. Using Z and C bits isn't likely to be much
' quicker or shorter in code anyway.
                
LMM_Skip        rdlong  acc,pc                          ' Determine what the next LMM instruction is
                add     pc,#4                           ' Skip the instruction itself
                and     acc,#$1FF                       ' Only keep the Cog address part
                
                sub     acc,#LMM_Ret                    ' LMM_Ret ?
                tjz     acc,#LMM_Fetch                  ' Yes - No parameters to skip
                
                add     acc,#LMM_Ret-LMM_Jmp            ' LMM_Jmp ?
                tjz     acc,#LMM_Add4_Fetch             ' Yes - Skip one parameter
                
                sub     acc,#LMM_Call-LMM_Jmp           ' LMM_Call ?
                tjz     acc,#LMM_Add4_Fetch             ' Yes - Skip one parameter
                
                add     acc,#LMM_Call-LMM_Execute       ' LMM_Execute ?
                tjz     acc,#LMM_Add4_Fetch             ' Yes - Skip one parameter
                
                add     pc,#8                           ' Otherwise, skip two parameters
                jmp     #LMM_Fetch

' *******************************************************************************************************
' *                                                                                                     *
' *             End of LMM Interpreter                                                                  *
' *                                                                                                     *
' *******************************************************************************************************

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *             Place LMM Program Variables / Registers / Global Constants / Fast Subroutines here      *
' *                                                                                                     *
' *******************************************************************************************************

myAdr           long    0
myReg           long    0

                fit     $1F0                    ' Ensure we do not exceed Cog maximum
                
DAT

' *******************************************************************************************************
' *                                                                                                     *
' *             Place LMM Program here                                                                  *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |             LMM Test Program 1 - Increment                                                          |
' `-----------------------------------------------------------------------------------------------------'

LmmTest1        jmp     #LMM_Load               ' myAdr := $7000
                long    myAdr
                long    $7000

LmmLoop1        add     myReg,#1                ' repeat
                wrlong  myReg,myAdr             '   long[ myAdr ] := ++myReg

                jmp     #LMM_Jmp
                long    @LmmLoop1

' .-----------------------------------------------------------------------------------------------------.
' |             LMM Test Program 2 - Increment within subroutine                                        |
' `-----------------------------------------------------------------------------------------------------'

LmmTest2        jmp     #LMM_Load               ' myAdr := $7000
                long    myAdr
                long    $7000

LmmLoop2        jmp     #LMM_Call               ' repeat
                long    @LmmSub2                '   LmmSub2

                jmp     #LMM_Jmp
                long    @LmmLoop2

LmmSub2         add     myReg,#1                ' PRI LmmSub2
                wrlong  myReg,myAdr             '   long[ myAdr ] := ++myReg

                jmp     #LMM_Ret

